import { READ_STATE } from "@src/const";
import type { CachedMetadata, TFile } from "obsidian";
import { getContext, setContext } from "svelte";
import { getSettingsContext } from "./settings.svelte";
import type BookTrackerPlugin from "@src/main";

export type FileMetadata = {
	file: TFile;
	frontmatter: Record<string, any>;
};

export type FileProperty = {
	file: TFile;
	value: any;
};

export const ALL_TIME = "ALL_TIME";

interface MetadataStore {
	get metadata(): FileMetadata[];
	filterYear: number | typeof ALL_TIME;
	filterMonth: number | typeof ALL_TIME;
	get filterYears(): number[];
	get filterMonths(): { label: string; value: number }[];

	destroy(): void;
}
export function createMetadata(plugin: BookTrackerPlugin): MetadataStore {
	const settings = getSettingsContext().settings;

	const initialMetadata: FileMetadata[] = [];

	for (const file of plugin.app.vault.getMarkdownFiles()) {
		const frontmatter =
			plugin.app.metadataCache.getFileCache(file)?.frontmatter ?? {};

		if (frontmatter[settings.statusProperty] !== READ_STATE) {
			continue;
		}

		initialMetadata.push({ file, frontmatter });
	}

	const thisYear = new Date().getFullYear();

	let metadata = $state(initialMetadata);
	let filterYear: number | typeof ALL_TIME = $state(thisYear);
	let filterMonth: number | typeof ALL_TIME = $state(ALL_TIME);
	const filteredMetadata = $derived.by(() => {
		console.log("here");
		return metadata.filter((f) => {
			// @ts-expect-error Moment is provided by Obsidian
			const endDate = moment(f.frontmatter[settings.endDateProperty]);
			if (filterYear !== ALL_TIME) {
				if (endDate.year() !== filterYear) {
					return false;
				}
			}

			if (filterMonth !== ALL_TIME) {
				if (endDate.month() !== filterMonth - 1) {
					return false;
				}
			}
			return true;
		});
	});

	const filterYears = $derived.by(() => {
		const years = new Set<number>();
		for (const f of metadata) {
			// @ts-expect-error Moment is provided by Obsidian
			const endDate = moment(f.frontmatter[settings.endDateProperty]);
			years.add(endDate.year());
		}
		return Array.from(years).sort((a, b) => a - b);
	});

	const filterMonths = $derived.by(() => {
		if (filterYear === ALL_TIME) {
			return [];
		}

		const months = [];
		for (const f of metadata) {
			// @ts-expect-error Moment is provided by Obsidian
			const endDate = moment(f.frontmatter[settings.endDateProperty]);

			if (endDate.year() !== filterYear) {
				continue;
			}

			months.push({
				label: endDate.format("MMMM"),
				value: endDate.month() + 1,
			});
		}

		return months
			.filter(
				(month, idx, self) =>
					idx === self.findIndex((m) => m.value === month.value)
			)
			.sort((a, b) => a.value - b.value);
	});

	function onChanged(file: TFile, _data: string, cache: CachedMetadata) {
		metadata = metadata.map((f) => {
			if (f.file.path === file.path) {
				return {
					...f,
					frontmatter: cache.frontmatter ?? {},
				};
			}
			return f;
		});
	}

	plugin.registerEvent(plugin.app.metadataCache.on("changed", onChanged));

	function onDeleted(file: TFile) {
		metadata = metadata.filter((f) => f.file.path !== file.path);
	}
	plugin.registerEvent(plugin.app.metadataCache.on("deleted", onDeleted));

	return {
		get metadata() {
			return filteredMetadata;
		},
		get filterYear() {
			return filterYear;
		},
		set filterYear(value) {
			filterYear = value;
		},
		get filterMonth() {
			return filterMonth;
		},
		set filterMonth(value) {
			filterMonth = value;
		},
		get filterYears() {
			return filterYears;
		},
		get filterMonths() {
			return filterMonths;
		},
		destroy() {
			plugin.app.metadataCache.off("changed", onChanged);
			plugin.app.metadataCache.off("deleted", onDeleted);
		},
	};
}

const METADATA_KEY = Symbol("metadata");

export function setMetadataContext(state: MetadataStore) {
	setContext(METADATA_KEY, state);
}

export function getMetadataContext(): MetadataStore {
	return getContext(METADATA_KEY) as MetadataStore;
}

function notEmpty(value: any): boolean {
	return value !== undefined && value !== null && value !== "" && value !== 0;
}

interface PropertyStore {
	get propertyData(): FileProperty[];
}

export function createPropertyStore(
	property: string,
	filter: (value: any) => boolean = notEmpty
): PropertyStore {
	const store = getMetadataContext();

	const propertyData = $derived(
		store.metadata
			.map((f) => ({ ...f, value: f.frontmatter[property] }))
			.filter((f) => (filter ? filter(f.value) : true))
	);

	return {
		get propertyData() {
			return propertyData;
		},
	};
}
